/* tex4ht.rc: version number                            -*- C++ -*-

   Copyright (C) 1998-2017 Christian Schenk

   This file is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2, or (at your
   option) any later version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file; if not, write to the Free Software
   Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. */

#include "t4ht-version.h"

#define VER_FILEDESCRIPTION_STR "TeX4ht - LaTeX and TeX for Hypertext"
#define VER_INTERNALNAME_STR "t4ht"
#define VER_ORIGINALFILENAME_STR "miktex-t4ht.exe"

#include "miktex/win/version.rc"
