/**
 * @file miktex-uptex.cpp
 * @author Christian Schenk
 * @brief MiKTeX upTeX
 *
 * @copyright Copyright © 2021-2022 Christian Schenk
 *
 * This file is free software; the copyright holder gives unlimited permission
 * to copy and/or distribute it, with or without modifications, as long as this
 * notice is preserved.
 */

#define MIKTEX_UPTEX_CPP
#include "miktex-uptex.h"
